def hello(str_name="gangster"):
    '''
    This is a sample function that returns a string.
    
    Parameters
    ----------
    str_name : str, optional
        [description], by default "gangster"
    
    Returns
    -------
    str
        The string returned.
    '''
    return "hi %s!" % (str_name) 


from abc import ABC, abstractmethod
import pickle
from pathlib import Path
from os import makedirs, chmod, remove, walk, rename
from os.path import isdir, isfile, getsize, join, islink
from shutil import copyfile, move, copytree, rmtree
from re import sub
from .logger import logger
from ..storage.storage import Storage


def safe_folder_path_str(path):
    path = str(path)
    path = sub('[^a-zA-Z0-9-_./]+', '', path)
    if len(path) > 0:
        path = path + "/"
        path = sub('[/]+', '/', path)
    return path


def safe_file_path_str(path):
    path = Path(path)
    path_folder = path.parent
    file_name = path.name
    path_folder = safe_folder_path_str(path_folder)
    file_name = sub('[^a-zA-Z0-9-_.]+', '', file_name)
    path = path_folder + file_name
    return path


def get_folder_size(start_path = '.'):
    # This function is copied from:
    # https://stackoverflow.com/questions/1392413/
    total_size = 0
    for dirpath, dirnames, filenames in walk(start_path):
        for f in filenames:
            fp = join(dirpath, f)
            # skip if it is symbolic link
            if not islink(fp):
                total_size += getsize(fp)
    return total_size


class StorageDisk(Storage):


    def __init__(self, root_path="/"):
        try:
            self.__storage_type = "DISK"
            root_path = str(root_path)
            assert root_path[0] == "/", "Root path should start with /."
            root_path = Path(root_path).resolve()
            assert isdir(root_path), "Root folder not found."
            self.__root_path_full = root_path
            self.__cd_full = root_path
            self.__cd = Path("/")
            self.__initialized = True
            logger.debug("Storage DISK initialized.")
        except Exception as e:
            self.__initialized = False
            logger.error("Initialization failed. " + str(e))


    def initialized(self):
        return self.__initialized


    def __path_expand(self, path):
        path = str(path)
        if len(path) == 0:
            path_full = self.__cd_full
        elif path[0] == "/":
            path_full = Path(str(self.__root_path_full) + path).resolve()
        else:
            path_full = (self.__cd_full / path).resolve()
        return path_full
    

    def __check_path_full(self, path_full):
        assert str(Path(path_full).resolve())\
            .startswith(str(self.__root_path_full)), \
            "Impossible to go beyond the root path."


    def get_type(self):
        try:
            assert self.__initialized, "Storage not initialized."
            logger.debug("Storage type: " + self.__storage_type)
            return self.__storage_type
        except Exception as e:
            logger.error("Failed to get the storage type. " + str(e))


    def cd(self, path):
        try:
            assert self.__initialized, "Storage not initialized."
            path = str(path)
            path_full = self.__path_expand(path)
            self.__check_path_full(path_full)
            assert isdir(path_full), "Current directory not found."
            self.__cd_full = path_full
            if path[0] == "/":
                self.__cd = Path(path).resolve()
            else:
                self.__cd = Path(str(self.__cd) + "/" + path).resolve()
            logger.debug("cd " + str(path) + ": True")
        except Exception as e:
            logger.error("cd failed. " + str(e))
    

    def pwd(self):
        try:
            assert self.__initialized, "Storage not initialized."
            #logger.debug("pwd full path: " + str(self.__cd_full))
            logger.debug("pwd: " + str(self.__cd))
            return str(self.__cd)
        except Exception as e:
            logger.error("pwd failed. " + str(e))


    def ls(self, path=""):
        try:
            assert self.__initialized, "Storage not initialized."
            path = str(path)
            path_full = self.__path_expand(path)
            self.__check_path_full(path_full)
            assert isdir(path_full), "Folder not found."
            output = [x.name for x in path_full.iterdir()]
            logger.debug("ls " + str(path) + ": " + " ".join(output))
            return output
        except Exception as e:
            logger.error("Failed to list objects inside the folder. " + str(e))


    def exists(self, path):
        try:
            assert self.__initialized, "Storage not initialized."
            path = str(path)
            path_full = self.__path_expand(path)
            self.__check_path_full(path_full)
            output = isdir(path_full) or isfile(path_full)
            logger.debug("exists " + str(path) + ": " + str(output))
            return output
        except Exception as e:
            logger.error("Failed to check the existence. " + str(e))


    def mkdir(self, path):
        try:
            assert self.__initialized, "Storage not initialized."
            path = str(path)
            path = safe_folder_path_str(path)
            path_full = self.__path_expand(path)
            self.__check_path_full(path_full)
            assert not isdir(path_full), "Directory already exists."
            makedirs(path_full)
            assert isdir(path_full), "Directory check failed."
            chmod(path_full, 0o777)
            logger.debug("mkdir " + str(path) + ": True")
        except Exception as e:
            logger.error("Failed to create the directory. " + str(e))  


    def upload(self, path_source, path_dest):
        try:
            assert self.__initialized, "Storage not initialized."
            path_source = str(path_source)
            path_dest = str(path_dest)
            path_dest = safe_file_path_str(path_dest)
            path_full = self.__path_expand(path_dest)
            self.__check_path_full(path_full)
            assert isfile(path_source), "Source file not found."
            assert not isfile(path_full), "Destination file already exists."
            assert not isdir(path_full), "Destination folder already exists."
            copyfile(path_source, path_full)
            assert isfile(path_full), "Destination file check failed."
            chmod(path_full, 0o777)
            logger.debug("upload " + str(path_dest) + ": True")
        except Exception as e:
            logger.error("Failed to upload. " + str(e))  


    def download(self, path_source, path_dest):
        try:
            assert self.__initialized, "Storage not initialized."
            path_source = str(path_source)
            path_dest = str(path_dest)
            path_source = safe_file_path_str(path_source)
            path_full = self.__path_expand(path_source)
            self.__check_path_full(path_full)
            assert isfile(path_full), "Source file not found."
            assert not isfile(path_dest), "Destination file already exists."
            assert not isdir(path_dest), "Destination folder already exists."
            copyfile(path_full, path_dest)
            assert isfile(path_dest), "Destination file check failed."
            chmod(path_dest, 0o777)
            logger.debug("download " + str(path_source) + ": True")
        except Exception as e:
            logger.error("Failed to download. " + str(e)) 


    def rm(self, path):
        try:
            assert self.__initialized, "Storage not initialized."
            path = str(path)
            path = safe_folder_path_str(path)
            path_full = self.__path_expand(path)
            self.__check_path_full(path_full)
            assert (isdir(path_full) or isfile(path_full)), \
                "File/folder not found."
            if isfile(path_full):
                remove(path_full)
            else:
                rmtree(path_full, ignore_errors=True)
            assert ((not isdir(path_full)) and (not isfile(path_full))), \
                "File/folder still exists."
            logger.debug("rm " + str(path) + ": True")
        except Exception as e:
            logger.error("Failed to remove the file/folder. " + str(e)) 


    def size(self, path):
        try:
            assert self.__initialized, "Storage not initialized."
            path = str(path)
            path = safe_file_path_str(path)
            path_full = self.__path_expand(path)
            self.__check_path_full(path_full)
            assert (isdir(path_full) or isfile(path_full)), \
                "File/folder not found."
            if isfile(path_full):
                output = getsize(path_full)
            else:
                output = get_folder_size(path_full)
            logger.debug("size " + str(path) + ": " + str(output))
            return output
        except Exception as e:
            logger.error("Failed to get the size. " + str(e))


    def upload_from_memory(self, variable, path):
        try:
            assert self.__initialized, "Storage not initialized."
            path = str(path)
            path = safe_file_path_str(path)
            path_full = self.__path_expand(path)
            self.__check_path_full(path_full)
            assert not isfile(path_full), "File already exists."
            assert not isdir(path_full), "Folder already exists."
            pickle.dump(obj=variable, file=open(path_full, "wb"))
            assert isfile(path_full), "File check failed."
            chmod(path_full, 0o777)
            logger.debug("upload_from_memory " + str(path) + ": True")
        except Exception as e:
            logger.error("Failed to upload. " + str(e))  


    def download_to_memory(self, path):
        try:
            assert self.__initialized, "Storage not initialized."
            path = str(path)
            path = safe_file_path_str(path)
            path_full = self.__path_expand(path)
            self.__check_path_full(path_full)
            assert isfile(path_full), "File not found."
            logger.debug("download_to_memory " + str(path) + ": True")
            return pickle.load(file=open(path_full, "rb"))
        except Exception as e:
            logger.error("Failed to download. " + str(e))  


    def rename(self, path_source, path_dest):
        try:
            assert self.__initialized, "Storage not initialized."
            path_source = str(path_source)
            path_dest = str(path_dest)
            path_source_full = self.__path_expand(path_source)
            self.__check_path_full(path_source_full)
            path_dest = safe_file_path_str(path_dest)
            path_dest_full = self.__path_expand(path_dest)
            assert path_source_full.parent == path_dest_full.parent, \
                "Different parent directories."
            self.__check_path_full(path_dest_full)
            assert (isfile(path_source_full) or isdir(path_source_full)), \
                "Source file/folder not found."
            assert not (isfile(path_dest_full) or isdir(path_dest_full)), \
                "Destination already exists."
            rename(path_source_full, path_dest_full)
            assert (isfile(path_dest_full) or isdir(path_dest_full)), \
                "Destination check failed."
            assert not (isfile(path_source_full) or isdir(path_source_full)), \
                "Source check failed."
            logger.debug("rename " + str(path_source) + \
                " --> " + str(path_dest))
        except Exception as e:
            logger.error("Failed to rename. " + str(e)) 


    def mv(self, path_source, path_dest):
        try:
            assert self.__initialized, "Storage not initialized."
            path_source = str(path_source)
            path_dest = str(path_dest)
            path_source_full = self.__path_expand(path_source)
            self.__check_path_full(path_source_full)
            path_dest = safe_file_path_str(path_dest)
            path_dest_full = self.__path_expand(path_dest)
            self.__check_path_full(path_dest_full)
            assert (isfile(path_source_full) or isdir(path_source_full)), \
                "Source file/folder not found."
            assert not (isfile(path_dest_full) or isdir(path_dest_full)), \
                "Destination already exists."
            move(path_source_full, path_dest_full)
            assert (isfile(path_dest_full) or isdir(path_dest_full)), \
                "Destination check failed."
            assert not (isfile(path_source_full) or isdir(path_source_full)), \
                "Source check failed."
            logger.debug("mv " + str(path_source) + \
                " --> " + str(path_dest))
        except Exception as e:
            logger.error("Failed to move. " + str(e)) 


    def cp(self, path_source, path_dest):
        try:
            assert self.__initialized, "Storage not initialized."
            path_source = str(path_source)
            path_dest = str(path_dest)
            path_source_full = self.__path_expand(path_source)
            self.__check_path_full(path_source_full)
            path_dest = safe_file_path_str(path_dest)
            path_dest_full = self.__path_expand(path_dest)
            self.__check_path_full(path_dest_full)
            assert (isfile(path_source_full) or isdir(path_source_full)), \
                "Source file/folder not found."
            assert not (isfile(path_dest_full) or isdir(path_dest_full)), \
                "Destination already exists."
            if isfile(path_source_full):
                copyfile(path_source_full, path_dest_full)
            else:
                copytree(path_source_full, path_dest_full)
            assert (isfile(path_dest_full) or isdir(path_dest_full)), \
                "Destination check failed."
            assert (isfile(path_source_full) or isdir(path_source_full)), \
                "Source check failed."
            logger.debug("cp " + str(path_source) + \
                " --> " + str(path_dest))
        except Exception as e:
            logger.error("Failed to copy. " + str(e)) 


    def append(self, path, content):
        try:
            assert self.__initialized, "Storage not initialized."
            path = str(path)
            path = safe_file_path_str(path)
            path_full = self.__path_expand(path)
            self.__check_path_full(path_full)
            assert isfile(path_full), "File not found."
            with open(path_full, "a") as f:
                f.write(content)
            logger.debug("append " + str(path) + ": " + str(content))
        except Exception as e:
            logger.error("Failed to append. " + str(e)) 


import logging
import logging.config
import yaml
import os
import re
from pathlib import Path
from os.path import dirname, realpath


if "LOG_LEVEL" not in os.environ:
    os.environ["LOG_LEVEL"] = "DEBUG"


def pattern():
    '''
    Pattern of the keyword used in YAML file to get values from the environment.
    
    Returns
    -------
    re.compile
        re.compile object.
    '''
    return re.compile(r'.*?\${(\w+)}.*?')


def constructor_env_variables(loader, node):
    '''
    Constructor of the env variables
    
    Parameters
    ----------
    loader : loader
        Loader.
    node : node
        Node.
    
    Returns
    -------
    loader object
        Loader object.
    '''
    value = loader.construct_scalar(node)
    match = pattern().findall(value)  
    if match:
        full_value = value
        for g in match:
            full_value = full_value.replace(
                f'${{{g}}}', os.environ.get(g, g)
            )
        return full_value
    return value


def get_logger(
    str_logger_name, 
    yaml_logging=Path(dirname(dirname(realpath(__file__))) + "/config/logging.yml")
):
    '''
    Get a custom logger.
    
    Parameters
    ----------
    str_logger_name : str
        The name of the logger.
    yaml_logging : str, optional
        Location of the YAML file for the logging configuration, by default "config/logging.yml"
    
    Returns
    -------
    logger
        The custom logger. 
    '''
    path_logging_yml = None
    if os.path.isfile(yaml_logging):
        path_logging_yml = Path(yaml_logging)
    if path_logging_yml is not None:
        tag = "!ENV"
        loader = yaml.SafeLoader
        loader.add_implicit_resolver(tag, pattern(), None)
        loader.add_constructor(tag, constructor_env_variables)
        with open(path_logging_yml, 'r') as stream:
            logger_config = yaml.load(stream, Loader=loader)
        logging.config.dictConfig(logger_config)
        logger = logging.getLogger(str_logger_name)
    else:
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - \
            [%(name)s] - [%(module)s] - %(funcName)s() - %(message)s')
        handler = logging.StreamHandler()
        handler.setLevel(level=logging.DEBUG)
        handler.setFormatter(formatter)
        logger = logging.getLogger(str_logger_name)
        logger.addHandler(handler)
        logger.propagate = False
    return logger


from abc import ABC, abstractmethod


class Storage(ABC):


    @abstractmethod
    def __init__(self):
        pass


    @abstractmethod
    def get_type(self):
        pass


    @abstractmethod
    def cd(self):
        pass


    @abstractmethod
    def pwd(self):
        pass


    @abstractmethod
    def ls(self):
        pass


    @abstractmethod
    def exists(self):
        pass


    @abstractmethod
    def mkdir(self):
        pass


    @abstractmethod
    def upload(self):
        pass


    @abstractmethod
    def download(self):
        pass


    @abstractmethod
    def rm(self):
        pass


    @abstractmethod
    def size(self):
        pass


    @abstractmethod
    def upload_from_memory(self):
        pass


    @abstractmethod
    def download_to_memory(self):
        pass


    @abstractmethod
    def rename(self):
        pass


    @abstractmethod
    def mv(self):
        pass


    @abstractmethod
    def cp(self):
        pass


    @abstractmethod
    def append(self):
        pass


import yaml
import os
from pathlib import Path
from os.path import dirname, realpath
import re
from .logger import logger


def pattern():
    '''
    Pattern of the keyword used in YAML file to get values from the environment.
    
    Returns
    -------
    re.compile
        re.compile object.
    '''
    return re.compile(r'.*?\${(\w+)}.*?')


def constructor_env_variables(loader, node):
    '''
    Constructor of the env variables
    
    Parameters
    ----------
    loader : loader
        Loader.
    node : node
        Node.
    
    Returns
    -------
    loader object
        Loader object.
    '''
    value = loader.construct_scalar(node)
    match = pattern().findall(value)  
    if match:
        full_value = value
        for g in match:
            full_value = full_value.replace(
                f'${{{g}}}', os.environ.get(g, g)
            )
        return full_value
    return value


def get_config(path_env_yaml=None):
    '''
    Get the configuration for the package.
    
    Parameters
    ----------
    path_env_yaml : str, optional
        Location of the YAML configuration file, by default None
    
    Returns
    -------
    dict
        A dictionary containing configuration parameters.
    '''

    if path_env_yaml is None:
        if "ENV_RUN" not in os.environ:
            logger.info("Environment not defined in os.environ, set DEVELOPMENT as default value.")
            os.environ["ENV_RUN"] = "DEVELOPMENT"
        str_env = os.environ["ENV_RUN"].lower()
        logger.info("Environment set to " + str_env + ".")
        path_env_yaml = Path(dirname(dirname(realpath(__file__))) + "/config/" + str_env + ".yml")

    if os.path.isfile(path_env_yaml):
        tag = "!ENV"
        loader = yaml.SafeLoader
        loader.add_implicit_resolver(tag, pattern(), None)
        loader.add_constructor(tag, constructor_env_variables)
        with open(path_env_yaml, 'r') as stream:
            dict_config = yaml.load(stream, Loader=loader)
    else:
        logger.warning("YAML file not found, loading default configuration.")
        dict_config = {
            "ENV": "DEFAULT", 
            "PING": "PONG", 
            "PATH_IN": "IN",
            "PATH_OUT": "OUT", 
            "PATH_SAVE": "SAVE"
        }

    return dict_config


dict_config = get_config()


from abc import ABC, abstractmethod
import pickle
from pathlib import Path
#from os import makedirs, chmod, remove, walk, rename
#from os.path import isdir, isfile, getsize, join, islink
#from shutil import copyfile, move, copytree, rmtree
from os.path import isdir, isfile
from re import sub
from boto.s3.connection import S3Connection
from boto.s3.key import Key
from io import BytesIO
from numpy import unique
from math import ceil
from .logger import logger
from ..storage.storage import Storage


def safe_folder_path_str(path):
    path = str(path)
    path = sub('[^a-zA-Z0-9-_./]+', '', path)
    if len(path) > 0:
        path = path + "/"
        path = sub('[/]+', '/', path)
    return path


def safe_file_path_str(path):
    path = Path(path)
    path_folder = path.parent
    file_name = path.name
    path_folder = safe_folder_path_str(path_folder)
    file_name = sub('[^a-zA-Z0-9-_.]+', '', file_name)
    path = path_folder + file_name
    return path


class StorageS3boto(Storage):


    def __init__(
        self, 
        host,
        port,
        access_key,
        secret_key, 
        bucket,
        calling_format,
        secure,
        root_path="/",
    ):
        try:
            self.__storage_type = "S3boto"
            root_path = str(root_path)
            assert len(root_path) > 0, "No root path provided."
            assert root_path[0] == "/", "Root path should start with /."
            root_path = str(Path(root_path).resolve())
            if root_path[-1] != "/":
                root_path = root_path + "/"
            self.__root_path_full = root_path
            self.__cd_full = root_path
            self.__cd = "/"
            self.__host = str(host)
            self.__port = int(port)
            self.__access_key = str(access_key)
            self.__secret_key = str(secret_key)
            self.__bucket = str(bucket)
            self.__calling_format = str(calling_format)
            if type(secure) == bool:
                self.__secure = secure 
            else:
                self.__secure = secure == "True" 

            self.__connection = S3Connection(
                host=self.__host,
                port=self.__port,
                aws_access_key_id=self.__access_key,
                aws_secret_access_key=self.__secret_key,
                calling_format=self.__calling_format,
                is_secure=self.__secure
            )

            assert self.__connection.lookup(self.__bucket) is not None, \
                "The bucket specified doesn't exists!"
            
            self.__connection_bucket = self.__connection\
                .get_bucket(self.__bucket)

            if len(self.__root_path_full) > 0: 
                k = Key(self.__connection_bucket)
                k.key = self.__rm_lead_slash(self.__root_path_full)
                assert k.exists(), "Root folder not found!"

            self.__initialized = True
            logger.debug("Storage DISK initialized.")

        except Exception as e:
            self.__initialized = False
            logger.error("Initialization failed. " + str(e))


    def initialized(self):
        return self.__initialized


    def __path_expand(self, path, bool_file=True):
        path = str(path)
        if len(path) == 0:
            assert not bool_file, "Not a file."
            path_full = self.__cd_full
        elif path[0] == "/":
            path_full = str(Path(self.__root_path_full + path).resolve())
            if not bool_file:
                path_full = path_full + "/"
        else:
            path_full = str((Path(self.__cd_full) / path).resolve())
            if not bool_file:
                path_full = path_full + "/"
        assert path_full.startswith(str(self.__root_path_full)), \
            "Impossible to go beyond the root path."
        return path_full
    

    def __rm_lead_slash(self, path):
        if path[0] == "/":
            return path[1:]
        else:
            return path
    

    def __exists(self, key):
        k = Key(self.__connection_bucket)
        k.key = key
        return k.exists()

    
    def __exists_parent(self, key):
        key_parent = str(Path("/" + key).parent)
        if key_parent[-1] != "/":
            key_parent = key_parent + "/"
        if key_parent == "/":
            return True
        else:
            k = Key(self.__connection_bucket)
            k.key = key_parent
            return k.exists()


    def get_type(self):
        try:
            assert self.__initialized, "Storage not initialized."
            logger.debug("Storage type: " + self.__storage_type)
            return self.__storage_type
        except Exception as e:
            logger.error("Failed to get the storage type. " + str(e))


    def cd(self, path):
        try:
            assert self.__initialized, "Storage not initialized."
            path = str(path)
            path_full = self.__path_expand(path, bool_file=False)
            assert self.__exists(path_full), "Current directory not found."
            self.__cd_full = path_full
            self.__cd = "/" + sub(self.__root_path_full, "", self.__cd_full)
            logger.debug("cd " + str(path) + ": True")
        except Exception as e:
            logger.error("cd failed. " + str(e))
    

    def pwd(self):
        try:
            assert self.__initialized, "Storage not initialized."
            output = self.__cd
            if output != "/" and output[-1] == "/":
                output = output[:-1]
            logger.debug("pwd: " + output)
            return output
        except Exception as e:
            logger.error("pwd failed. " + str(e))


    def ls(self, path=""):
        try:
            assert self.__initialized, "Storage not initialized."
            path = str(path)
            path_full = self.__path_expand(path, bool_file=False)
            path_full_4_s3 = self.__rm_lead_slash(path_full) 
            if len(path_full_4_s3) > 0:
                assert self.__exists(path_full_4_s3), "Folder not found."
                iterable = self.__connection_bucket.list(prefix=path_full_4_s3)
                output = [x.name for x in iterable]
                output = [x.name for x in iterable if x.name != path_full_4_s3]
                output = [sub("^"+path_full_4_s3+"|/.*$", "", x) for x in output]
            else:
                iterable = self.__connection_bucket.list()
                output = [x.name for x in iterable]
                output = [sub("/.*$", "", x) for x in output]
            logger.debug("ls " + str(path) + ": " + " ".join(output))
            return unique(output)
        except Exception as e:
            logger.error("Failed to list objects inside the folder. " + str(e))


    def exists(self, path):
        try:
            assert self.__initialized, "Storage not initialized."
            path = str(path)
            path_full = self.__path_expand(path, bool_file=True)
            path_full = self.__rm_lead_slash(path_full)
            k = Key(self.__connection_bucket)
            k.key = path_full
            if k.exists():
                output = True
            else:
                k.key = path_full + "/"
                output = k.exists()
            logger.debug("exists " + str(path) + ": " + str(output))
            return output
        except Exception as e:
            logger.error("Failed to check the existence. " + str(e))


    def mkdir(self, path):
        try:
            assert self.__initialized, "Storage not initialized."
            path = str(path)
            path = safe_folder_path_str(path)
            path_full = self.__path_expand(path, bool_file=False)
            path_full_4_s3 = self.__rm_lead_slash(path_full)
            assert not self.__exists(path_full_4_s3), \
                "Directory already exists."
            assert self.__exists_parent(path_full_4_s3), \
                "Parent folder not found"
            k = self.__connection_bucket.new_key(path_full_4_s3)
            k.set_contents_from_string('')
            assert self.__exists(path_full_4_s3), "Directory check failed."
            logger.debug("mkdir " + str(path) + ": True")
        except Exception as e:
            logger.error("Failed to create the directory. " + str(e))  


    def upload(self, path_source, path_dest):
        try:
            assert self.__initialized, "Storage not initialized."
            path_source = str(path_source)
            path_dest = str(path_dest)
            path_dest = safe_file_path_str(path_dest)
            path_full = self.__path_expand(path_dest, bool_file=True)
            path_full_4_s3 = self.__rm_lead_slash(path_full)
            assert self.__exists_parent(path_full_4_s3), \
                "Parent folder not found."
            assert isfile(path_source), "Source file not found."
            assert not self.__exists(path_full_4_s3), \
                "Destination file already exists."
            assert not self.__exists(path_full_4_s3 + "/"), \
                "Destination folder already exists."
            k = self.__connection_bucket.new_key(path_full_4_s3)
            with open(path_source, "rb") as fp:
                k.set_contents_from_file(fp)
            assert self.__exists(path_full_4_s3), \
                "Destination file check failed."
            logger.debug("upload " + str(path_dest) + ": True")
        except Exception as e:
            logger.error("Failed to upload. " + str(e))  


    def download(self, path_source, path_dest):
        try:
            assert self.__initialized, "Storage not initialized."
            path_source = str(path_source)
            path_dest = str(path_dest)
            path_source = safe_file_path_str(path_source)
            path_full = self.__path_expand(path_source, bool_file=True)
            path_full_4_s3 = self.__rm_lead_slash(path_full)
            assert self.__exists(path_full_4_s3), "Source file not found."
            assert not isfile(path_dest), "Destination file already exists."
            assert not isdir(path_dest), "Destination folder already exists."
            with open(path_dest, "wb") as fp:
                self.__connection_bucket\
                    .get_key(path_full_4_s3)\
                    .get_contents_to_file(fp)
            assert isfile(path_dest), "Destination file check failed."
            logger.debug("download " + str(path_source) + ": True")
        except Exception as e:
            logger.error("Failed to download. " + str(e)) 


    def rm(self, path):
        try:
            assert self.__initialized, "Storage not initialized."
            path = str(path)
            path = safe_folder_path_str(path)
            path_full = self.__path_expand(path, bool_file=True)
            path_full_4_s3 = self.__rm_lead_slash(path_full)
            assert len(path_full_4_s3) > 0, "Nothing to remove."
            if self.__exists(path_full_4_s3):
                self.__connection_bucket.delete_key(path_full_4_s3)
                assert not self.__exists(path_full_4_s3), \
                    "File/folder still exists."
            iterable = self.__connection_bucket\
                .list(prefix=path_full_4_s3+"/")
            output = [x.name for x in iterable]
            for k in output:
                self.__connection_bucket.delete_key(k)
                assert not self.__exists(k), "File/folder still exists."
            logger.debug("rm " + str(path) + ": True")
        except Exception as e:
            logger.error("Failed to remove the file/folder. " + str(e)) 


    def size(self, path):
        try:
            assert self.__initialized, "Storage not initialized."
            path = str(path)
            path = safe_file_path_str(path)
            path_full = self.__path_expand(path, bool_file=True)
            path_full_4_s3 = self.__rm_lead_slash(path_full)
            if self.__exists(path_full_4_s3):
                output = self.__connection_bucket.get_key(path_full_4_s3).size
            elif self.__exists(path_full_4_s3 + "/"):
                iterable = self.__connection_bucket\
                    .list(prefix=path_full_4_s3 + "/")
                output = sum([x.size for x in iterable])
            else:
                raise ValueError( "File/folder not found.")
            logger.debug("size " + str(path) + ": " + str(output))
            return output
        except Exception as e:
            logger.error("Failed to get the size. " + str(e)) 


    def upload_from_memory(self, variable, path, bool_bin=False):
        try:
            assert self.__initialized, "Storage not initialized."
            path = str(path)
            path = safe_file_path_str(path)
            path_full = self.__path_expand(path, bool_file=True)
            path_full_4_s3 = self.__rm_lead_slash(path_full)
            assert not self.__exists(path_full_4_s3), "File already exists."
            assert not self.__exists(path_full_4_s3 + "/"), "Folder already exists."
            if bool_bin:
                content=variable
            else:
                content = pickle.dumps(variable)
            mp = self.__connection_bucket.initiate_multipart_upload(path_full_4_s3)
            chunk_size = 5242880
            source_size = len(content)
            chunk_count = int(ceil(source_size / float(chunk_size)))
            for i in range(chunk_count):
                offset = chunk_size * i
                int_bytes = min(chunk_size, source_size - offset)
                fp = BytesIO(content[offset:offset+int_bytes])
                mp.upload_part_from_file(fp, part_num=i+1)
            mp.complete_upload()
            assert self.__exists(path_full_4_s3), "File check failed."
            logger.debug("upload_from_memory " + str(path) + ": True")
        except Exception as e:
            logger.error("Failed to upload. " + str(e))  


    def download_to_memory(self, path, bool_bin=False):
        try:
            assert self.__initialized, "Storage not initialized."
            path = str(path)
            path = safe_file_path_str(path)
            path_full = self.__path_expand(path, bool_file=True)
            path_full_4_s3 = self.__rm_lead_slash(path_full)
            assert self.__exists(path_full_4_s3), "File not found."
            with BytesIO() as b:
                k = self.__connection_bucket.get_key(path_full_4_s3)
                k.get_file(b)
                b.seek(0)
                if bool_bin:
                    output = b.read()
                else:
                    output = pickle.loads(b.read())
            logger.debug("download_to_memory " + str(path) + ": True")
            return output
        except Exception as e:
            logger.error("Failed to download. " + str(e))  


    def rename(self, path_source, path_dest):
        try:
            assert self.__initialized, "Storage not initialized."
            path_source = str(path_source)
            path_source_full = self.__path_expand(path_source, bool_file=True)
            path_source_full_4_s3 = self.__rm_lead_slash(path_source_full)
            path_dest = str(path_dest)
            path_dest = safe_file_path_str(path_dest)
            path_dest_full = self.__path_expand(path_dest, bool_file=True)
            path_dest_full_4_s3 = self.__rm_lead_slash(path_dest_full)
            assert Path(path_dest_full_4_s3).parent \
                == Path(path_source_full_4_s3).parent, \
                "Different parent directories."
            if self.__exists(path_source_full_4_s3) \
                and not self.__exists(path_dest_full_4_s3):
                self.__connection_bucket.copy_key(
                    path_dest_full_4_s3, 
                    self.__bucket, 
                    path_source_full_4_s3
                )
                self.__connection_bucket.delete_key(path_source_full_4_s3)
            else:
                assert self.__exists(path_source_full_4_s3+"/") \
                    and not self.__exists(path_dest_full_4_s3+"/"), \
                    "Source not found or destination already exists."
                iterable = self.__connection_bucket\
                    .list(prefix=path_source_full_4_s3+"/")
                array_sources = [x.name for x in iterable]
                array_dests = [sub("^"+path_source_full_4_s3, \
                    path_dest_full_4_s3, x) \
                    for x in array_sources]
                for item_dest in array_dests:
                    assert not self.__exists(item_dest), \
                        "Destination already exists."
                for j, item_source in enumerate(array_sources):
                    self.__connection_bucket.copy_key(
                        array_dests[j], 
                        self.__bucket, 
                        item_source
                    )
                    self.__connection_bucket.delete_key(item_source)
                    assert self.__exists(array_dests[j]), \
                        "Destination check failed."
                    assert not self.__exists(item_source), \
                        "Source check failed."
            logger.debug("rename " + str(path_source) + \
                " --> " + str(path_dest))
        except Exception as e:
            logger.error("Failed to rename. " + str(e)) 


    def mv(self, path_source, path_dest):
        try:
            assert self.__initialized, "Storage not initialized."
            path_source = str(path_source)
            path_source_full = self.__path_expand(path_source, bool_file=True)
            path_source_full_4_s3 = self.__rm_lead_slash(path_source_full)
            path_dest = str(path_dest)
            path_dest = safe_file_path_str(path_dest)
            path_dest_full = self.__path_expand(path_dest, bool_file=True)
            path_dest_full_4_s3 = self.__rm_lead_slash(path_dest_full)
            if self.__exists(path_source_full_4_s3) \
                and not self.__exists(path_dest_full_4_s3):
                self.__connection_bucket.copy_key(
                    path_dest_full_4_s3, 
                    self.__bucket, 
                    path_source_full_4_s3
                )
                self.__connection_bucket.delete_key(path_source_full_4_s3)
            else:
                assert self.__exists(path_source_full_4_s3+"/") \
                    and not self.__exists(path_dest_full_4_s3+"/"), \
                    "Source not found or destination already exists."
                iterable = self.__connection_bucket\
                    .list(prefix=path_source_full_4_s3+"/")
                array_sources = [x.name for x in iterable]
                array_dests = [sub("^"+path_source_full_4_s3, \
                    path_dest_full_4_s3, x) \
                    for x in array_sources]
                for item_dest in array_dests:
                    assert not self.__exists(item_dest), \
                        "Destination already exists."
                for j, item_source in enumerate(array_sources):
                    self.__connection_bucket.copy_key(
                        array_dests[j], 
                        self.__bucket, 
                        item_source
                    )
                    self.__connection_bucket.delete_key(item_source)
                    assert self.__exists(array_dests[j]), \
                        "Destination check failed."
                    assert not self.__exists(item_source), \
                        "Source check failed."
            logger.debug("mv " + str(path_source) + \
                " --> " + str(path_dest))
        except Exception as e:
            logger.error("Failed to move. " + str(e)) 


    def cp(self, path_source, path_dest):
        try:
            assert self.__initialized, "Storage not initialized."
            path_source = str(path_source)
            path_source_full = self.__path_expand(path_source, bool_file=True)
            path_source_full_4_s3 = self.__rm_lead_slash(path_source_full)
            path_dest = str(path_dest)
            path_dest = safe_file_path_str(path_dest)
            path_dest_full = self.__path_expand(path_dest, bool_file=True)
            path_dest_full_4_s3 = self.__rm_lead_slash(path_dest_full)
            if self.__exists(path_source_full_4_s3) \
                and not self.__exists(path_dest_full_4_s3):
                self.__connection_bucket.copy_key(
                    path_dest_full_4_s3, 
                    self.__bucket, 
                    path_source_full_4_s3
                )
            else:
                assert self.__exists(path_source_full_4_s3+"/") \
                    and not self.__exists(path_dest_full_4_s3+"/"), \
                    "Source not found or destination already exists."
                iterable = self.__connection_bucket\
                    .list(prefix=path_source_full_4_s3+"/")
                array_sources = [x.name for x in iterable]
                array_dests = [sub("^"+path_source_full_4_s3, \
                    path_dest_full_4_s3, x) \
                    for x in array_sources]
                for item_dest in array_dests:
                    assert not self.__exists(item_dest), \
                        "Destination already exists."
                for j, item_source in enumerate(array_sources):
                    self.__connection_bucket.copy_key(
                        array_dests[j], 
                        self.__bucket, 
                        item_source
                    )
                    assert self.__exists(array_dests[j]), \
                        "Destination check failed."
                    assert self.__exists(item_source), \
                        "Source check failed."
            logger.debug("cp " + str(path_source) + \
                " --> " + str(path_dest))
        except Exception as e:
            logger.error("Failed to copy. " + str(e)) 


    def append(self, path, content):
        try:
            logger.warning("Not the most efficient implementation, improve it!")
            assert self.__initialized, "Storage not initialized."
            assert type(content) == str, \
                "content should be a string"
            path = str(path)
            path = safe_file_path_str(path)
            path_full = self.__path_expand(path, bool_file=True)
            path_full = self.__rm_lead_slash(path_full)
            k = Key(self.__connection_bucket)
            k.key = path_full
            assert k.exists(), "File not found."
            content_old = self.download_to_memory(path=path)
            assert type(content_old) == str, \
                "It is only possible to append to strings!"
            content_new = content_old + content
            self.rm(path)
            self.upload_from_memory(variable=content_new, path=path)
            logger.debug("append " + str(path) + ": " + str(content))
        except Exception as e:
            logger.error("Failed to append. " + str(e)) 


from setuptools import setup, find_packages
import os
import sys


if sys.version_info[0] < 3:
    with open('README.md') as f:
        long_description = f.read()
else:
    with open('README.md', encoding='utf-8') as f:
        long_description = f.read()

with(open("version.txt", "r")) as f:
    version = f.read()

with(open("requirements.txt", "r")) as f:
    requirements = f.read()


setup(
    name='sdaab',
    version=version,
    description='Simple Data Abstraction to access different types of storages.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='ngshya',
    author_email='ngshya@gmail.com',
    url='https://github.com/ngshya/sdaab',
    license='GPLv3',
    packages=find_packages(exclude=["tests"]),
    install_requires=requirements,
    include_package_data=True,
)